#encoding: utf-8
"""
Author: Mika Hämäläinen, University of Helsinki
url: https://mikakalevi.com

This is the code used to produce the SemFi database

INSTRUCTIONS AT THE END OF THE FILE

"""
import codecs, json
from mikatools import *
from collections import defaultdict
from uralicNLP import uralicApi
import sqlite3
from scipy.stats import zscore
import numpy

files = ["arcs.04.txt","arcs.02.txt","arcs.03.txt","arcs.01.txt"]

poses = [u'N', u'Num', u'V', u'Interj', u'A', u'Pron', u'Adv', u'Adp',u'C']
relations = [u'xsubj', u'voc', u'intj', u'csubj', u'amod', u'preconj', u'cc', u'cop', u'conj', u'gobj', u'ROOT', u'advmod', u'gsubj', u'appos', u'nommod', u'nsubj', u'rcmod', u'xcomp', u'acomp', u'adpos,adpos', u'advcl', u'nsubj-cop', u'dep', u'infmod', u'comparator', u'iccomp', u'rel', u'xsubj-cop', u'dobj', u'csubj-cop', u'parataxis', u'num', u'nn', u'nommod-own', u'name', u'partmod', u'number',  u'compar', u'adpos', u'quantmod', u'ccomp']

pos_readings = {u"V": "+V+Act+InfA+Sg+Lat", u"Num": "+Num+Sg+Nom", u"N":"+N+Sg+Nom", u"Interj" : "+Interj", u"A":"+A+Sg+Nom", u"Adv":"+Adv", u"Adp":"+Adp+Po"}

def check_multiword(word, pos):
	if type(word) is unicode:
		try:
			word = word.encode("utf-8")
		except:
			return False
	if num_there(word):
		return False
	if "|" in word:
		if pos == "C" or pos == "Pron" or pos == "Interj" or pos == "Adp" or pos == "Adv" or pos == "Num":
			return False
		parts = word.split("|")
		last = parts.pop()
		for part in parts:
			a = uralicApi.analyze(part, "fin")
			if len(a) == 0:
				return False
			for an in a:
				if "+Punct" in an[0]:
					return False
		return __check_pos(last, pos)
	return __check_pos(word, pos)

def num_there(s):
    return any(i.isdigit() for i in s)

def __check_pos(word, pos):
	analysis = uralicApi.analyze(word, "fin")
	for a in analysis:
		res = __check_analysis(pos, a[0])
		if res:
			return True
	return False

def __check_analysis(pos, a):
	if pos == u"C":
		if a.endswith("+CC") or a.endswith("+CS"):
			return True
		else:
			 return False
	elif pos == u"Pron":
		if a.endswith("+Nom") and "+Pron" in a:
			return True
		else:
			return False
	else:
		if a.endswith(pos_readings[pos]):
			return True
		else:
			return False

def poses_and_relations():
	poses= SafeDict(int)
	relations = SafeDict(int)
	for file in files:
		f = codecs.open(file, "r", encoding="utf-8")
		i = 0
		for line in f:
			words = line.split(" ")
			for word in words:
				parts = word.split("/")
				pos = parts[2]
				relation = parts[4]
				poses[pos] += 1
				relations[relation] += 1
	json_dump(poses, "poses.json")
	json_dump(relations, "relations.json")

def build_semfi():
	for pos in poses:
		nested_dict = lambda: defaultdict(nested_dict)
		results_dict = nested_dict()
		for file in files:
			f = codecs.open(file, "r", encoding="utf-8")
			i = 0
			for line in f:
				lemma1, pos1, lemma2, pos2, rel = process_line(line)
				if lemma1 is None:
					continue
				if pos1 != pos:
					continue
				rel_dict = results_dict[lemma1]
				pos_d = rel_dict[rel][pos2]
				if lemma2 not in pos_d:
					pos_d[lemma2] = 0
				pos_d[lemma2] += 1
		json_dump(results_dict, "results/" + pos + ".json")



def process_line(line):
	lemma, reading = line.split("\t",1)
	words = reading.split(" ")
	word1 = words[0]
	word2 = words[1]

	parts1 = word1.split("/")
	lemma1 = parts1[1]
	pos1 = parts1[2]

	parts2 = word2.split("/")
	lemma2 = parts2[1]
	pos2 = parts2[2]

	rel1 = parts1[4]
	rel2 = parts2[4]
	if lemma == lemma1:
		rel = rel1
	else:
		rel = rel2
	if pos1 not in poses or pos2 not in poses or rel not in relations or lemma1 == u"" or lemma2 == u"":
		return None, None, None, None, None
	return lemma1, pos1, lemma2, pos2, rel


#poses_and_relations()
def print_poses_relations():
	poses = json_load("poses.json")
	relations = json_load("relations.json")
	print [name for name, age in poses.items() if age > 1000]
	print [name for name, age in relations.items() if age > 1000]


def filter_result_keys():
	for pos in poses:
		data = json_load("results/" + pos + ".json")
		words = list(data.keys())
		print pos
		for word in words:
			r = check_multiword(word, pos)
			if not r:
				del data[word]
		print "Starting to dump", pos
		json_dump(data, "results/" + pos + ".json")
		print "Done!"


def filter_results_values():
	for pos in poses:
		data = json_load("results/" + pos + ".json")
		print pos
		for word in data:
			d = data[word]
			for relation in d:
				da = d[relation]
				for p in da:
					w = da[p]
					words = list(w.keys())
					for wor in words:
						if not check_multiword(wor, p):
							del w[wor]
		print "Starting to dump", pos
		json_dump(data, "results/" + pos + ".json")
		print "Done!"

def add_quotes(lista):
	result = []
	for l in lista:
		if type(l) == str or type(l) == unicode:
			result.append("\""+ l + "\"")
		else:
			result.append(str(l))
	return result

def fill_db_words():
	total_freq = 134915948.0
	conn = sqlite3.connect('semfi.db')
	c = conn.cursor()
	for pos in poses:
		data = json_load("results/" + pos + ".json")
		for word in data:
			f = get_frequencies(data[word])
			values = [word +"_"+ pos, word, pos, f, f/total_freq, int("|" in word)]
			c.execute("INSERT INTO words(id, word, pos, frequency, relative_frequency, compound) VALUES ("+ ",".join(add_quotes(values)) + ")")


	conn.commit()
	conn.close()

def get_frequencies(word_data):
	freq = 0
	for relation in word_data.values():
		for words in relation.values():
			for word in words:
				freq += words[word]
	return freq

def fill_db_relations():
	conn = sqlite3.connect('semfi.db')
	c = conn.cursor()
	for pos in poses:
		data = json_load("results/" + pos + ".json")
		z_data = json_load("results/" + pos + "_zscore.json")
		for word, word_data in data.iteritems():
			z_i = z_data[word]
			for relation, relation_data in word_data.iteritems():
				z_r = z_i[relation]
				for rel_pos, words in relation_data.iteritems():
					total = float(sum(words.values()))
					z_info = z_r[rel_pos]
					for rel_word in words:
						zs = z_info[rel_word]
						if numpy.isnan(zs):
							zs = 0
						values = [word + "_" + pos, rel_word + "_" + rel_pos, relation, words[rel_word], words[rel_word]/total, zs]
						c.execute("INSERT INTO relations(word1, word2, relation_name, frequency, relative_frequency, zscore) VALUES ("+ ",".join(add_quotes(values)) + ")")


	conn.commit()
	conn.close()


def z_score():
	for pos in poses:
		print pos
		data = json_load("results/" + pos + ".json")
		for word, word_data in data.iteritems():
			for relation, relation_data in word_data.iteritems():
				for rel_pos, words in relation_data.iteritems():
					keys = list(words.keys())
					vals = []
					for key in keys:
						vals.append(words[key])
					z_scores = zscore(vals)
					for i in range(len(keys)):
						words[keys[i]] = z_scores[i]
		print "dumping"
		json_dump(data, "results/" + pos + "_zscore.json")








### SEMFI FROM SCRATCH ###
# Uncomment and run these to build the jsons and filter them with omorfi
"""
build_semfi()
uralicApi.download("fin")
filter_result_keys()
filter_results_values()
z_score()
"""


### FROM JSONS TO SQLITE ###
#You need to create the tables manually

"""
CREATE TABLE "words" ( `id` TEXT NOT NULL, `word` TEXT, `pos` TEXT, `frequency` NUMERIC, `relative_frequency` NUMERIC, `compound` INTEGER, PRIMARY KEY(`id`) )
CREATE TABLE "relations" ( `word1` TEXT NOT NULL, `word2` TEXT NOT NULL, `relation_name` TEXT NOT NULL, `frequency` NUMERIC, `relative_frequency` NUMERIC, `zscore` NUMERIC, FOREIGN KEY(`word1`) REFERENCES `words`(`id`), FOREIGN KEY(`word2`) REFERENCES `words`(`id`) )
"""

#After running the following methods, use sqlite to remove words with 0 frequency
"""
fill_db_words()
fill_db_relations()
"""

### EXAMPLE OF THE STRUCTURE OF THE INPUT FILES ###
# These files are downloadable from Turku University NLP group website
"""
ovat	ovat/ovat/V/PRS_Pl3|VOICE_Act|TENSE_Prs|MOOD_Ind|OTHER_UNK/ROOT/0 ,/,/Punct/_/punct/1	4
ovat	ovat/ovat/V/PRS_Pl3|VOICE_Act|TENSE_Prs|MOOD_Ind|OTHER_UNK/ROOT/0 ././Punct/_/punct/1	4
"""

